from wdltest.wdltest import Wdltest
import os
import sys
import logging
import logging.config
import signal
import time
import subprocess

# Create the Logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Create the Handler for logging data to a file
logger_handler = logging.StreamHandler(sys.stdout)
logger_handler.setLevel(logging.DEBUG)

# Create a Formatter for formatting the log messages
logger_formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')

# Add the Formatter to the Handler
logger_handler.setFormatter(logger_formatter)

# Add the Handler to the Logger
logger.addHandler(logger_handler)
logger.info('Configured logger handler')

def sigterm_handler(_signo, _stack_frame):
    cromwell.stop()
    sys.exit(0)
signal.signal(signal.SIGTERM, sigterm_handler)
signal.signal(signal.SIGINT, sigterm_handler)

def hello():
    return "wdltest hello"

#def server_testrun():
#    return Wdltest().run()

def local_testrun():
    return Wdltest().localrun()



from wdltest.cromwell_handler import CromwellHandler
from wdltest.test_configuration import TestConfiguration
from wdltest.test_runner import TestRunner

import logging
import os
import configparser

class Wdltest(object):

    def __init__(self, testConfigFile = os.path.dirname(__file__) + "/test_module_config.json", configFile = os.path.dirname(__file__) + "/wdltest.cfg", index = -1):
        self.logger = logging.getLogger(__name__)
        self.logger.info('Reading config')
        config = self.getConfig(configFile)
        testConfig = TestConfiguration(testConfigFile).getConfiguration()

        # Preparing cromwell
        self.logger.info('Preparing cromwell handler')
        self.cromwell = CromwellHandler(config)
        self.testRunner = TestRunner(testConfig, self.cromwell, index = index)

    def getConfig(self, configFile = os.path.dirname(__file__) + "/wdltest.cfg"):
        config = configparser.ConfigParser()
        config.read(configFile)
        return config

    def run(self):
        exitCode = self.testRunner.run()
        self.cromwell.stop()
        return exitCode

    def localrun(self):
        exitCode = self.testRunner.run()
        self.cromwell.stop()
        return exitCode

    def stop(self):
        self.cromwell.stop()


import json
import re
import os
import time
import logging
import subprocess
import threading
import time

class TestRunner(object):

    def __init__(self, configuration, cromwell, index = -1):
        self.logger = logging.getLogger(__name__)
        self.configuration = configuration
        self.cromwell = cromwell
        self.index = index
        self.exitCode = 0
        maxThreads = 1
        if "threads" in self.configuration:
            maxThreads = self.configuration["threads"]
        self.sema = threading.Semaphore(value = maxThreads)

    def run(self):
        self.logger.debug("Cromwell started")
        testIndex = 0
        self.logger.debug("Index " + str(self.index))
        
        threads = list()
        for testJson in self.configuration["tests"]:
            thread = threading.Thread(target = self.runTest, args=(testIndex, testJson))
            threads.append(thread)
            thread.start()
            time.sleep(2)
            testIndex = testIndex + 1
        for thread in threads:
            thread.join()
        return self.exitCode

    def runTest(self, testIndex, testJson):
        self.sema.acquire()
        exitCode = 0
        if self.index != -1 and self.index != testIndex:
            self.logger.info("Excluding test with index " + str(testIndex))
            return 0
        try:
            self.cromwell.submitJob(self.configuration["wdl"], testJson["inputs"])
            status = "Started"
            start = time.time()
            while status != "Failed" and status != "Succeeded":
                time.sleep(1)
                previousStatus = status
                status = self.cromwell.getStatus()
                if status != previousStatus:
                    start = time.time()
                    if previousStatus != "Started": print()
                diff = int(time.time() - start)
                print("Cromwell job status " + status + " " + str(diff) + "s                     ", end ="\r")
            print()
            for condition in testJson["conditions"]:
                try:
                    if self.cromwell.getPathToOutput(condition["file"]) == 'missing':
                        self.logger.error("[ERROR] Test '" + condition["name"] + "' failed with message: no " + condition["file"] + " file in workflow outputs")
                        exitCode = 1
                    else:
                        bashCommand = condition["command"].replace("$file", self.cromwell.getPathToOutput(condition["file"]))
                        process = subprocess.Popen(bashCommand, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        output, error = process.communicate()
                        output = output.decode("utf-8").strip() + error.decode("utf-8").strip()
                        returnCode = process.poll()
                        if returnCode == 0:
                            self.logger.info("[PASSED] Test '" + condition["name"] + "'")
                        else:
                            if "error_message" in condition.keys():
                                self.logger.error("[ERROR] Test '" + condition["name"] + "' failed with message: " + condition["error_message"])
                            else:
                                self.logger.error("[ERROR] Test '" + condition["name"] + "' failed")
                            exitCode = 1
                
                except Exception as e:
                    self.logger.error("[ERROR] Test '" + condition["name"] + "' failed because file " + condition["file"] + " does not exists. " + e)
                    exitCode = 1
                    
        except Exception as e:
            exitCode = 1
            print("ERROR " + str(e))
            self.cromwell.stop()   
            self.sema.release()
        if (exitCode != 0 and self.exitCode == 0):
            self.exitCode = exitCode       
        self.sema.release()



import subprocess
import requests
import os
import logging
import threading
import socket
import random
import time
import json
import regex
import getpass ## for username
from shutil import copyfile
from datetime import datetime

class CromwellHandler(object):

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.config = config
        self.checkJava()
        self.downloadCromwell()
        self.mode = self.config['cromwell']['mode']
        if(self.mode == "server"):
            self.port = self.getPort()
            self.localhost = "127.0.0.1"
            cromwellThread = threading.Thread(target = self.startCromwell)
            cromwellThread.start()
            self.waitForCromwell()

    def checkJava(self):
        self.logger.debug('Checking if java is installed')
        bashCommand = "java -version"
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output, error = process.communicate()
        output = output.decode("utf-8").strip() + error.decode("utf-8").strip()
        if(not (("vm" in output) or ("VM" in output))):
            raise Exception("JVM is not installed")

    def downloadCromwell(self):
        dir = self.config['paths']['dir']
        if(not os.path.isdir(dir)):
            os.mkdir(dir)
            os.chmod(dir, 0o776)
        path = dir + "/cromwell.jar"
        url = self.config['cromwell']['url']
        if(not os.path.isfile(path)):
            self.logger.debug('Downloading cromwell')
            request = requests.get(url, allow_redirects=True)
            open(path, 'wb').write(request.content)
            os.chmod(path, 0o666)
        configSourcePath = os.path.dirname(__file__) + "/cromwell.cfg"
        configDestinationPath = dir + "/cromwell.cfg" 
        if(not os.path.isfile(configDestinationPath)):
            copyfile(configSourcePath, configDestinationPath)
            os.chmod(configDestinationPath, 0o666)

    def getPort(self):
        for i in range(0, 19):
            port = random.randint(8000,8999)
            if self.isPortAvailable(port):
                return port
        raise Exception("Could not find free port to allocate for Cromwell")

    def isPortAvailable(self, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((self.localhost,port))
        if result == 0:
            return False
        else:
            return True
        sock.close()

    def startCromwell(self):
        bashCommand = "java -Dconfig.file=" + self.config['paths']['dir'] + " -Dwebservice.port=" + str(self.port) + " -jar " + self.config['paths']['dir'] + "/cromwell.jar server"
        self.cromwellProcess = subprocess.Popen(bashCommand.split(), cwd=self.config['paths']['dir'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def getCromwellUrl(self):
        return "http://" + self.localhost + ":" + str(self.port)

    def waitForCromwell(self):
        self.logger.debug("Waiting for cromwell server")
        url = self.getCromwellUrl() + "/engine/v1/status"
        for i in range(0, 29):
            time.sleep(1)
            try:
                resp = requests.get(url)
                return
            except:
                pass
        raise Exception("Could not connect to Cromwell")

    def submitJob(self, wdl, inputs):
        self.logger.debug("Submmitting job in " + self.mode + " mode")
        if self.mode == "server":
            multipart_form_data = {
                'workflowSource': ("workflow.wdl", open( wdl, "rb")),
                'workflowInputs': ("inputs.json", json.dumps(inputs)),
            }
            response = requests.post(self.getCromwellUrl() + "/api/workflows/v1", files=multipart_form_data)
            self.job = json.loads(response.content.decode("utf-8"))["id"]
            self.logger.debug("Submitted job " + self.job)
        if self.mode == "run":
            self.basePath = self.config['paths']['dir'] + "/" + getpass.getuser() + "/" + os.path.basename(os.path.dirname(wdl))
            if(not os.path.isdir(self.basePath)):
                os.makedirs(self.basePath)
                os.chmod(self.basePath, 0o776)
            self.runPath = self.basePath + "/" + datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            if(not os.path.isdir(self.runPath)):
                os.makedirs(self.runPath)
                os.chmod(self.runPath, 0o776)
            self.inputPath = self.runPath + "/inputs.json"
            self.logPath = self.runPath + "/cromwell-execution.log"

            self.returnCode = -1
            self.logger.debug("PATH: " + self.runPath)
            with open(self.inputPath, "w") as inputs_file:
                print(json.dumps(inputs), file=inputs_file)
            bashCommand = "java -Dconfig.file=" + self.config['paths']['dir'] + "/cromwell.cfg" + " -jar " + self.config['paths']['dir'] + "/cromwell.jar run " + wdl + " --inputs " + self.inputPath
            self.cromwellProcess = subprocess.Popen(bashCommand.split(), cwd=self.runPath, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.log = ""
            self.status = "Running"
            with open(self.logPath, "w") as log_file:
                for line in iter(self.cromwellProcess.stdout.readline, ""):
                    printline = line.decode("utf-8").replace('\n', '')
                    if "to Done" in printline:
                        print(printline)
                    print(printline, file=log_file)
                    self.log = self.log + line.decode("utf-8").replace('\n', '')
                    if "workflow finished with status" in printline:
                        self.status = printline.split()[-1].replace("'", '').replace('.',' ')
                    if line == b'' and self.cromwellProcess.poll() is not None:
                        break
            #print (self.status)
            if "Succeeded" in self.status:
                #print("in Succeeded")
                #self.logger.debug(self.log)
                pattern = regex.compile(r'\{(?:[^{}]|(?R))*\}')
                jsons = pattern.findall(self.log)
                #self.logger.debug(jsons)
                for jsonitem in jsons:
                    #self.logger.debug(jsonitem)
                    if jsonitem.find('"outputs":') != -1:
                        #self.logger.debug("IN")
                        rawoutputs = json.loads(jsonitem)
                #self.logger.debug(self.outputs)
                rawoutputs = rawoutputs["outputs"]
                self.outputs = dict()
                for key in rawoutputs:
                    #self.logger.debug("key: " + key)
                    try:
                        
                        #self.logger.debug("key: " + key.split('.')[-1])
                    
                        self.outputs[key.split('.')[-1]] = rawoutputs[key]
                    except Exception as e:
                        self.logger.debug("error: " + str(e))
                self.cromwellProcess.stdout.close()
                self.returnCode = self.cromwellProcess.wait()
                self.logger.debug("Finished run")
                self.logger.info("Path to log " + self.logPath)
            else:
                #print("not succeeded")
                self.outputs = dict()
                self.cromwellProcess.stdout.close()
                self.logger.info("Workflow failed")
                self.logger.info("Path to log " + self.logPath)
                self.returnCode = 1
            #print(self.outputs)

    def getStatus(self):
        if self.mode == "server":
            response = requests.get(self.getCromwellUrl() + "/api/workflows/v1/" + self.job + "/status")
            status = json.loads(response.content.decode("utf-8"))
            if status["status"] == "fail":
                return "Submitted unregistered"
            return status["status"]
        if self.mode == "run":
            if self.returnCode == -1:
                return "Running"
            if self.returnCode == 0:
                return "Succeeded"
            if self.returnCode > 0:
                return "Failed"

    def getMetadata(self):
        response = requests.get(self.getCromwellUrl() + "/api/workflows/v1/" + self.job + "/metadata")
        return json.loads(response.content.decode("utf-8"))

    def getOutputs(self):
        response = requests.get(self.getCromwellUrl() + "/api/workflows/v1/" + self.job + "/outputs")
        return json.loads(response.content.decode("utf-8"))

    def getPathToOutput(self, desiredOutput):
        if self.mode == "server":
            outputs = self.getOutputs()["outputs"]
            #print(outputs)
            for key in outputs.keys():
                workflow, output = os.path.splitext(key)
                if(output.replace('.', '') == desiredOutput):
                    return outputs[key]
            raise Exception("Could not find path to output: " + desiredOutput)
        if self.mode == "run":
            if desiredOutput in self.outputs:
                output = self.outputs[desiredOutput]
                if "list" in type(output).__name__:
                    return output[0]
                return output
            else:
                return 'missing'
            

    def stop(self):
        if self.mode == "server":
            self.cromwellProcess.terminate()
            self.cromwellProcess.kill()


import json
import re
import os

class TestConfiguration(object):

    def __init__(self, path):
        with open(path, 'r') as json_file:
            json_text = json_file.read()
            variables = re.findall(r"\$\{[A-Za-z_-]+\}", json_text)
            for variable in variables:
                varname = variable.replace('$','').replace('{','').replace('}','')
                try:
                    json_text = json_text.replace(variable,os.environ[varname])
                except: raise Exception("Variable " + varname + " is undefined")
            self.testConfig = json.loads(json_text)
        for test in self.testConfig["tests"]:
            if "bcoCheck" in test:
                test["conditions"] = test["conditions"] + self.getDefaultBcoConditions()
            if "stdoutCheck" in test:
                test["conditions"] = test["conditions"] + self.getDefaultStdoutConditions()

    def getConfiguration(self):
        return self.testConfig

    def getDefaultBcoConditions(self):
        bcoConditions = [{
                    "file":"bco",
                    "name":"Bco exists",
                    "error_message":"Bco does not exist",
                    "command":"echo $file"
                },
                {
                    "file":"bco",
                    "name":"Provenance domain exists and is not empty",
                    "error_message":"Provenance domain not found in bco or is empty",
                    "command":"grep -q -m1 provenance_domain $file && jq -e 'if (.provenance_domain | length) == 0 then false else true end' $file"
                },
                {
                    "file":"bco",
                    "name":"Execution domain exists and is not empty",
                    "error_message":"Execution domain not found in bco or is empty",
                    "command":"grep -q -m1 execution_domain $file && jq -e 'if (.execution_domain | length) == 0 then false else true end' $file"
                },
                {
                    "file":"bco",
                    "name":"Parametric domain exists and is not empty",
                    "error_message":"Parametric domain not found in bco or is empty",
                    "command":"grep -q -m1 parametric_domain $file && jq -e 'if (.parametric_domain | length) == 0 then false else true end' $file"
                },
                {
                    "file":"bco",
                    "name":"Description domain exists and is not empty",
                    "command":"grep -q -m1 descripcion_domain $file && jq -e 'if (.description_domain | length) == 0 then false else true end' $file"
                }]
        return bcoConditions

    def getDefaultStdoutConditions(self):
        stdoutConditions = [    
                {
                    "file":"stdout_log",
                    "name":"Stdout exits",
                    "command":"test -f $file"
                }]
        return stdoutConditions



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wdltest",
    version="1.0.5",
    author="Marcin Piechota",
    author_email="piechota.marcin@gmail.com",
    description="Package for testing wdl workflows",
    #long_description=long_description,
    #long_description_content_type="text/reStructuredText",
    url="https://github.com/marpiech/wdltest",
    packages=setuptools.find_packages(),
    package_data={'wdltest': ['*.cfg']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['threaded','regex'],
    python_requires='>=3.6',
    scripts=['bin/wdltest'],
    test_suite='nose.collector',
    tests_require=['nose>=1.0'],
    setup_requires=['nose>=1.0'],
)
